import numpy as np  # Arrays
import pandas as pd  # Dataframes
from scipy.stats import median_abs_deviation  # MAD tests
from scipy.signal import find_peaks
from sklearn.cluster import KMeans


# Function to scale data
def _normalize_data(data):
    return (data - np.min(data)) * 100 / (np.max(data) - np.min(data))


# Derive RMSSD
def _get_rmssd(inputpeaks, samplingrate, minpeaks):
    if len(inputpeaks) > minpeaks:
        ibi = np.array([x - inputpeaks[i - 1] for i, x in enumerate(inputpeaks)][1:])  # Interbeat Intervals
        ibi_ms = ibi * 1000 / samplingrate  # Convert to ms (1000)
        sd = np.diff(ibi_ms)  # Successive differences in IBI
        ssd = np.square(sd)  # Square of successive differences
        mssd = np.mean(ssd)  # Mean square of successive differences
        rmssd = np.sqrt(mssd)  # Root mean square of successive differences
        return rmssd
    else:
        return np.nan


# Derive SDNN
def _get_sdnn(inputpeaks, samplingrate, minpeaks):
    if len(inputpeaks) > minpeaks:
        ibi = np.array([x - inputpeaks[i - 1] for i, x in enumerate(inputpeaks)][1:])  # NN Intervals
        ibi_ms = ibi * 1000 / samplingrate  # Convert to ms (1000)
        sdnn = np.std(ibi_ms)  # Standard deviation of NN intervals
        return sdnn
    else:
        return np.nan


# Derive BPM
def _get_bpm(inputpeaks, samplingrate, minpeaks):
    if len(inputpeaks) > minpeaks:
        act_window = ((inputpeaks[-1] - inputpeaks[0]) / samplingrate)
        bpm = ((len(inputpeaks) - 1) * 60) / act_window
        return bpm
    else:
        return np.nan


# Outlier detection
def _outlier_detect(inputdata, inputpeaks, properties, minpeaks, minwindow, samplingrate, mad, ibimad, bpmrange,
                    rmssdrange):
    if len(inputpeaks) < minpeaks:
        return True, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan
    else:
        # Get BPM/RMSSD
        bpm = _get_bpm(inputpeaks=inputpeaks, samplingrate=samplingrate, minpeaks=minpeaks)
        rmssd = _get_rmssd(inputpeaks=inputpeaks, samplingrate=samplingrate, minpeaks=minpeaks)

        # Gather data
        base_height = inputdata[inputpeaks] - properties['prominences']  # Baseline
        prominences = properties['prominences'] - np.mean(properties['prominences'])  # Mean center prominences
        heights = properties['peak_heights'] - np.mean(properties['peak_heights'])  # Mean center heights
        ibi = np.array([x - inputpeaks[i - 1] for i, x in enumerate(inputpeaks)][1:])  # Interbeat intervals
        ibi = ibi - np.mean(ibi)  # Mean-center

        # Check window length
        actwindowwidth = ((inputpeaks[-1] - inputpeaks[0]) / samplingrate)

        # Detect prominence outliers
        prom_mad_units = median_abs_deviation(prominences)  # Calculate MAD units
        prom_outliers = [np.greater(prominences, mad * prom_mad_units) |
                         np.less(prominences, -mad * prom_mad_units)]  # Detect outliers

        # Detect height outliers
        height_mad_units = median_abs_deviation(heights)  # Calculate MAD units
        height_outliers = [np.greater(heights, mad * height_mad_units) |
                           np.less(heights, -mad * height_mad_units)]  # Detect  outliers

        # Detect successive difference outliers
        ibi_mad_units = median_abs_deviation(ibi)  # Calculate MAD units
        ibi_outliers = [np.greater(ibi, ibimad * ibi_mad_units) |
                        np.less(ibi, -ibimad * ibi_mad_units)]  # Detect  outliers

        # Height thresholds
        height_thresholds = [(-mad * height_mad_units) + np.mean(properties['peak_heights']),
                             (mad * height_mad_units) + np.mean(properties['peak_heights'])]

        # Prominence thresholds
        prom_low_threshold = base_height + np.mean(properties['prominences']) - (mad * prom_mad_units)
        prom_high_threshold = base_height + np.mean(properties['prominences']) + (mad * prom_mad_units)
        prom_thresholds = [prom_low_threshold, prom_high_threshold]

        outlier_check = np.any(prom_outliers) | np.any(height_outliers) | np.any(ibi_outliers) |\
                        (actwindowwidth < minwindow) | (bpm < bpmrange[0]) | (bpm > bpmrange[1]) | \
                        (rmssd < rmssdrange[0]) | (rmssd > rmssdrange[1])

        return outlier_check, prom_thresholds, height_thresholds, prom_outliers, height_outliers, ibi_outliers, \
               base_height


# Function to run peak detection on a window.
def _get_peaks(inputdata, distance=200, prominence=50, k=1):
    peaks, properties = find_peaks(inputdata, distance=distance, height=0, prominence=prominence, width=0)

    if (len(peaks) >= 3) & (k > 1):
        clusteringdata = np.empty([len(peaks), 3])  # Generate feature matrix
        clusteringdata[:] = np.NaN
        clusteringdata[:, 0] = properties['widths'][:]
        clusteringdata[:, 1] = properties['peak_heights'][:]
        clusteringdata[:, 2] = properties['prominences'][:]

        kmeans = KMeans(n_clusters=k).fit(clusteringdata)  # Cluster data in P, R, and T waves
        widthcen = kmeans.cluster_centers_[:, 0]  # Get centroid of width feature

        # Check if width centroids are close (within 5). If so, use other parameter
        if np.abs(np.diff(np.sort(widthcen)[0:2])[0]) < 5:  # If less than 5, use height instead
            widthidx = np.argsort(widthcen)[0:2]  # Get indices of two lowest widths
            heightcen = kmeans.cluster_centers_[:, 2]  # Prominence centroids
            maxheight = np.max(heightcen[widthidx])  # Maximum prominence centroid for two lowest widths
            peakcluster = np.where(heightcen == maxheight)[0][0]  # Which label has higher prominence
        else:
            peakcluster = np.where(widthcen == np.min(widthcen))[0][0]  # Which label has lower width

        correctpeaks = kmeans.labels_ == peakcluster
        peaks = peaks[correctpeaks]  # Replace original peak/property values
        for dictkey in properties.keys():
            properties[dictkey] = properties[dictkey][correctpeaks]

    if len(peaks) > 3:
        baseheight = inputdata[peaks] - properties['prominences']  # Baseline
        properties['prominences'][0] = properties['peak_heights'][0] - baseheight[1]
        properties['prominences'][-1] = properties['peak_heights'][-1] - baseheight[-2]
        # The nature of random sampling windows means the baseline value may be off the mark. As such, the baseline for
        # the first peak is estimated using neighboring peak. This requires at least 3 peaks.

    return peaks, properties


def _moving_window(inputdata, windowwidth, windowmovement, samplingrate, numpeaksneeded, mad, ibimad, minamplitude,
                   minwindow, bpmrange, rmssdrange, mindistance, clusters):
    data_frame_cols = ['Time', 'BPM', 'CleanedBPM', 'RMSSD', 'CleanedRMSSD', 'SDNN', 'CleanedSDNN']
    rapiddata = pd.DataFrame(columns=data_frame_cols)

    for section in range(0, len(inputdata) - int(windowwidth * samplingrate) + 1, int(windowmovement * samplingrate)):
        exptime = section / samplingrate  # Note time

        # RAPID HRV PIPELINE
        segment = inputdata[section:section + (
                windowwidth * samplingrate)]  # Take initial window
        normed_data = _normalize_data(data=segment)  # Scale so min = 0, max = 100
        peaks, properties = _get_peaks(inputdata=normed_data, distance=mindistance, prominence=minamplitude, k=clusters)

        bpm = _get_bpm(inputpeaks=peaks, samplingrate=samplingrate, minpeaks=numpeaksneeded)  # HR
        rmssd = _get_rmssd(inputpeaks=peaks, samplingrate=samplingrate, minpeaks=numpeaksneeded)  # HRV
        sdnn = _get_sdnn(inputpeaks=peaks, samplingrate=samplingrate, minpeaks=numpeaksneeded)

        outliers, _, _, _, _, _, _ = \
            _outlier_detect(inputdata=normed_data, inputpeaks=peaks, properties=properties, minpeaks=numpeaksneeded,
                            minwindow=minwindow, samplingrate=samplingrate, mad=mad, ibimad=ibimad, bpmrange=bpmrange,
                            rmssdrange=rmssdrange)

        if outliers:
            cleaned_bpm = np.nan
            cleaned_rmssd = np.nan
            cleaned_sdnn = np.nan
        else:
            cleaned_bpm = bpm
            cleaned_rmssd = rmssd
            cleaned_sdnn = sdnn

        rapiddata = rapiddata.append(
            pd.DataFrame(data=[[exptime, bpm, cleaned_bpm, rmssd, cleaned_rmssd, sdnn, cleaned_sdnn]],
                         columns=data_frame_cols))

    return rapiddata, clusters


# Go over with rolling window
def extract_heart(inputdata, resampledrate, ecgclustering=False, windowwidth=10, windowmovement=None, minamplitude=50,
                  outliermethod='moderate', mad=None, ibimad=None, bpmrange=None, rmssdrange=None, numpeaksneeded=None,
                  minwindow=None, mindistance=None):
    """Function to extract heart rate (BPM) and variability (RMSSD) from heart data

    Parameters: inputdata (list): preprocessed cardiac data to be analysed.
    resampledrate (int): sampling rate in Hz. If data has been upsampled from preprocessing, make sure this. If no
    upsampling has been performed at preprocessing, use original sampling rate.
    ecgclustering (bool): whether to use k-means clustering to discern P, R, and T waves in the ECG signal. This
    overrides minamplitude and mindistance, setting them to 10 and 1 respectively. For clean ECG data that contains
    atypical morphologies (e.g. T wave amplitude > R wave amplitude), this may be useful. Optional (default=False)
    windowwidth (int): width of sliding window (in seconds) for which measures are extracted. Optional (default=10)
    windowmovement (int): how much the sliding window is shifted by (in seconds). Optional (default=windowwidth)
    minamplitude (int): the minimum amplitude before a peak is detected as a heart beat. Lower values = greater
    sensitivity, though at the risk of false detections. For PPG data, it is recommended to lower this to ~30.
    Optional (default=50)
    outliermethod (string): this will accept the arguments 'none','liberal','moderate','conservative' to determine
    the stringency of outlier rejection. 'liberal' will likely result in the least amount of cleaning, but risks
    introducing noise. 'conservative' will likely result in a larger amount of excluded data, but is less likely to
    introduce noise. Optional (default 'moderate')
    mad (int): the threshold (in median absolute deviation units) to detect whether peak height and prominences may be
    noise/outliers. Optional (if outliermethod == 'moderate': default=5)
    ibimad (int): the threshold (in median absolute deviation units) for detecting outliers based on peak intervals.
    Optional (if outliermethod == 'moderate': default=5)
    bpmrange (list): sets the lower and upper limits for deciding a window is an outlier based on BPM. Optional
    (if oulitermethod == 'moderate', default = [30, 190])
    rmssdrange (list): sets the lower and upper limits for deciding a window is an outlier based on RMSSD. Optional
    (if outliermethod == 'moderate', default = [5, 262])
    numpeaksneeded (int): the number of peaks needed for successful BPM/RMSSD extraction. Applies to both cleaned and
    uncleaned outputs. All input floored at 3. Optional (default = round((windowwidth / 5) + 2))
    minwindow (int): the minimum time (s) from first to last peak before a window is classified as an outlier. Optional
    (default=windowwidth/2)
    mindistance (int): minimum distance (in samples) required between peaks. Optional
    (default = round(samplingrate / 4) ≈ 250ms)

    Returns:
    dictionary:output. ['data'] is Dataframe containing experiment time (starting from 0), Uncleaned BPM, Cleaned BPM,
    Uncleaned RMSSD, and Cleaned RMSSD. ['features'] contains parameters from extract_heart.

   """

    if outliermethod == 'liberal':
        if bpmrange is None:
            bpmrange = [20, 200]
        if rmssdrange is None:
            rmssdrange = [0, 300]
        if mad is None:
            mad = 7
        if ibimad is None:
            ibimad = 7
    elif outliermethod == 'moderate':
        if bpmrange is None:
            bpmrange = [30, 190]
        if rmssdrange is None:
            rmssdrange = [5, 262]
        if mad is None:
            mad = 5
        if ibimad is None:
            ibimad = 5
    elif outliermethod == 'conservative':
        if bpmrange is None:
            bpmrange = [40, 180]
        if rmssdrange is None:
            rmssdrange = [10, 200]
        if mad is None:
            mad = 4  # Note: has been changed to 4
        if ibimad is None:
            ibimad = 4
    else:
        if (bpmrange is None) | (rmssdrange is None) | (mad is None) | (ibimad is None):
            raise ValueError('extract_heart requires either outliermethod or outlier-specific arguments (bpmrange,'
                             'rmssdrange, madrange, and ibimadrange). Is the outliermethod argument spelt correctly?')

    if windowmovement is None:
        windowmovement = windowwidth

    if minwindow is None:
        minwindow = windowwidth / 2

    if numpeaksneeded is None or numpeaksneeded < 3:
        if int(np.round((windowwidth / 5) + 2)) < 3:
            numpeaksneeded = 3
        else:
            numpeaksneeded = int(np.round((windowwidth / 5) + 2))

    if mindistance is None:
        mindistance = round(resampledrate / 4)
    else:
        mindistance = 500

    if ecgclustering:
        minamplitude = 5
        mindistance = 1
        clusters = 3
    else:
        clusters = 1

    data, k = _moving_window(inputdata, windowwidth, windowmovement, resampledrate, numpeaksneeded, mad, ibimad,
                             minamplitude, minwindow, bpmrange, rmssdrange, mindistance, clusters)

    output = {
        'data': pd.DataFrame(columns=['Time', 'BPM', 'CleanedBPM', 'RMSSD', 'CleanedRMSSD', 'SDNN', 'CleanedSDNN']),
        'features': {'samplingrate': resampledrate, 'windowwidth': windowwidth, 'windowmovement': windowmovement,
                     'numpeaksneeded': numpeaksneeded, 'mad': mad, 'ibimad': ibimad, 'minamplitude': minamplitude,
                     'minwindow': minwindow, 'bpmrange': bpmrange, 'rmssdrange': rmssdrange, 'mindistance': mindistance,
                     'ecgclustering': clusters}}

    output['data'] = output['data'].append(data, ignore_index=True)

    return output



import numpy as np  # Arrays
from scipy.signal import butter, filtfilt, savgol_filter  # Signal processing
from scipy.interpolate import splrep, splev


# Define highpass filter
def _butter_highpass(cutoff, fs, order):
    nyq = 0.5 * fs
    cutoff = cutoff / nyq
    b, a = butter(N=order, Wn=cutoff, btype='highpass', output='ba')
    return b, a


# Apply butterworth filter
def _butter_highpass_filter(data, cutoff, fs, order=5):
    b, a = _butter_highpass(cutoff=cutoff, fs=fs, order=order)
    y = filtfilt(b=b, a=a, x=data)
    return y


# Define lowpass filters
def _butter_lowpass(cutoff, fs, order):
    nyq = 0.5 * fs
    cutoff = cutoff / nyq
    b, a = butter(N=order, Wn=cutoff, btype='lowpass', output='ba')
    return b, a


# Apply butterworth filter
def _butter_lowpass_filter(data, cutoff, fs, order=5):
    b, a = _butter_lowpass(cutoff=cutoff, fs=fs, order=order)
    y = filtfilt(b=b, a=a, x=data)
    return y


def preprocess(inputdata, samplingrate, resample=1000, highpass=0.5, lowpass=None, polynomial=3, smooth=100):
    """Function to high pass filter and smooth cardiac data

    Parameters:
    inputdata (list): cardiac data to be preprocessed
    samplingrate (int): sampling rate in Hz
    resample(int): specify frequency in Hz. Uses cubic spline interpolation for resampling. Input 0 if not wanted.
    Optional (if samplingrate < resample, default=1000; else default=0)
    highpass (float/int): Optional (default=0.5)
    lowpass (float/int): Optional (default=None)
    polynomial(int): order of polynomial for savitzky-golay smoothing. Optional (default=3)
    smooth(int): window in ms for for savitzky-golay smoothing. Input None if not wanted. Optional (default=100)

    Returns:
    list:smoothed_data

   """

    if np.sum(np.isnan(inputdata)):
        raise ValueError(f'inputdata contains nan. preprocess requires no nans.')
    else:
        output = inputdata

    # First: upsample if specified
    if samplingrate < resample:
        if not resample % samplingrate == 0:
            raise ValueError(f'resample resolution (default=1000,used={resample}) must be divisible by original '
                             f'sampling rate ({samplingrate})')
        expansion = int(resample / samplingrate)
        samples = list(range(0, len(inputdata) * expansion, expansion))  # pretend samples are further apart in time
        tck = splrep(samples, inputdata, s=0, k=3)
        xnew = list(range(0, len(inputdata) * expansion))
        output = splev(xnew, tck, der=0)

    else:
        resample = samplingrate  # If no upsampling applied, redefine resample as original Hz for filtering

    # Second: highpass filter data
    if highpass:
        output = _butter_highpass_filter(data=output, cutoff=highpass, fs=resample)

    # Third: lowpass filter data (IF USING)
    if lowpass:
        output = _butter_lowpass_filter(data=output, cutoff=lowpass, fs=resample)

    # Smoothing with Savitsky-Golay filter
    if smooth:
        if smooth > 0:
            smoothing_window = int(np.round(samplingrate / (samplingrate / smooth)))
            if (smoothing_window % 2) == 0:  # If smoothing window is even (Sav-Gov requires odd)
                smoothing_window = smoothing_window + 1
            output = savgol_filter(output, window_length=smoothing_window, polyorder=polynomial)

    return output



from rapidhrv.analyze import _outlier_detect, _normalize_data, _get_peaks
import numpy as np
import matplotlib.pyplot as plt  # Plotting


# Visualize data
def _time_series_visualization(inputframe, ylim=None):
    if ylim is None:
        ylim = [30, 180, 2, 262]
    propscreened = (np.count_nonzero(np.isnan(inputframe['data']['BPM'])) * 100) / len(inputframe['data']['BPM'])
    propoutliers = (np.count_nonzero(np.isnan(inputframe['data']['CleanedBPM'])) * 100) / len(
        inputframe['data']['CleanedBPM']) - propscreened

    fig, axs = plt.subplots(4, 1)
    fig.set_size_inches(18, 10)
    fig.suptitle('Heart Rate and Variability Time Series', fontsize=25, fontweight='bold')

    # Text/labels
    boxprop = dict(boxstyle='round', pad=0.4, facecolor='khaki', alpha=0.7)
    rejection_text = f'{round(propoutliers + propscreened, 2)}% of windows excluded: \n' \
                     f'{round(propscreened, 2)}% = insufficient peaks \n' \
                     f'{round(propoutliers, 2)}% = outliers'
    plt.gcf().text(0.1, 0.92, s=rejection_text, fontsize=11, bbox=boxprop)
    axs[0].set_ylabel('BPM', fontsize=15)
    axs[0].xaxis.set_ticklabels([])
    axs[1].set_ylabel('BPM (cleaned)', fontsize=15)
    axs[1].xaxis.set_ticklabels([])
    axs[2].set_ylabel('RMSSD', fontsize=15)
    axs[2].xaxis.set_ticklabels([])
    axs[3].set_ylabel('RMSSD (cleaned)', fontsize=15)
    axs[3].set_xlabel('Time (seconds)', fontsize=20)

    # Grid
    for axisnum in range(0, 4):
        axs[axisnum].grid(b=True, axis='x', color='darkgrey')
        axs[axisnum].spines['bottom'].set_linewidth(2)
        axs[axisnum].spines['left'].set_linewidth(2)
        axs[axisnum].spines['top'].set_linewidth(0)
        axs[axisnum].spines['right'].set_linewidth(0)

    # HR Axes
    axs[0].set_xlim(inputframe['data']['Time'].iloc[0] - 5, inputframe['data']['Time'].iloc[-1] + 5)
    axs[0].set_ylim(ylim[0] - 20, ylim[1] + 20)
    axs[1].set_xlim(inputframe['data']['Time'].iloc[0] - 5, inputframe['data']['Time'].iloc[-1] + 5)
    axs[1].set_ylim(ylim[0] - 20, ylim[1] + 20)

    # HRV Axes
    axs[2].set_xlim(inputframe['data']['Time'].iloc[0] - 5, inputframe['data']['Time'].iloc[-1] + 5)
    axs[2].set_ylim(ylim[2] - 20, ylim[3] + 20)
    axs[3].set_xlim(inputframe['data']['Time'].iloc[0] - 5, inputframe['data']['Time'].iloc[-1] + 5)
    axs[3].set_ylim(ylim[2] - 20, ylim[3] + 20)

    # Drop NaNs and sort by cleaning
    rawoutput = inputframe['data'][['Time', 'BPM', 'RMSSD']].dropna()
    cleanedoutput = inputframe['data'][['Time', 'CleanedBPM', 'CleanedRMSSD']].dropna()

    # Plotting
    axs[0].plot(rawoutput['Time'], rawoutput['BPM'], color='red', linewidth=3, alpha=0.3)
    axs[0].scatter(rawoutput['Time'], rawoutput['BPM'], s=4, color='red')
    axs[1].plot(cleanedoutput['Time'], cleanedoutput['CleanedBPM'], color='red', linewidth=3, alpha=0.3)
    axs[1].scatter(cleanedoutput['Time'], cleanedoutput['CleanedBPM'], s=4, color='red')
    axs[2].plot(rawoutput['Time'], rawoutput['RMSSD'], color='purple', linewidth=3, alpha=0.3)
    axs[2].scatter(rawoutput['Time'], rawoutput['RMSSD'], s=4, color='purple')
    axs[3].plot(cleanedoutput['Time'], cleanedoutput['CleanedRMSSD'], color='purple', linewidth=3, alpha=0.3)
    axs[3].scatter(cleanedoutput['Time'], cleanedoutput['CleanedRMSSD'], s=4, color='purple')

    return fig


# Class to make figure interactive
class Visualize:
    """Class to plot uncleaned and cleaned heart rate and variability time series. The user can click on points in
    the time series. This will plot the specific window, and let the user know whether said window was cleaned.

     Parameters:
         inputdata (list): cardiac data that was analysed, ideally preprocessed
         inputframe (pandas DataFrame): analysed data, returned from extract_heart
         ylim (list): minimum and maximum values for BPM and RMSSD ya axes [BPM min, BPM max, RMSSD min, RMSSD max].
         These are adjusted slightly by RapidHRV (+-20). Optional (default = [30, 180, 2, 262])

     Returns:
     matplotlib object

    """

    def __init__(self, inputdata, inputframe, ylim=None):
        if ylim is None:
            ylim = [30, 180, 2, 262]
        self.processeddata = inputdata

        self.bpm = inputframe['data']['BPM']
        self.rmssd = inputframe['data']['RMSSD']
        self.exptime = inputframe['data']['Time']

        self.samplingrate = inputframe['features']['samplingrate']
        self.windowwidth = inputframe['features']['windowwidth']
        self.numpeaksneeded = inputframe['features']['numpeaksneeded']
        self.mad = inputframe['features']['mad']
        self.ibimad = inputframe['features']['ibimad']
        self.minamplitude = inputframe['features']['minamplitude']
        self.minwindow = inputframe['features']['minwindow']
        self.bpmrange = inputframe['features']['bpmrange']
        self.rmssdrange = inputframe['features']['rmssdrange']
        self.mindistance = inputframe['features']['mindistance']
        self.k = inputframe['features']['ecgclustering']

        fig = _time_series_visualization(inputframe=inputframe, ylim=ylim)
        fig.canvas.mpl_connect('button_press_event', self.onclick)

    def onclick(self, event):
        self.window_plot(event.xdata)  # Once user clicks on graph, pass to window_plot

    def window_plot(self, xclick):
        time_array = np.asarray(self.exptime)
        idx = int((np.abs(time_array - xclick)).argmin())
        section = int(time_array[idx] * self.samplingrate)
        timepoint = round(self.exptime.iloc[idx], 2)

        # Get plotting data
        normed_data = _normalize_data(self.processeddata[section:section + (self.windowwidth * self.samplingrate)])

        # Initialize plot
        fig, axarr = plt.subplots()
        segtime = np.arange(0, self.windowwidth,
                            self.windowwidth / (self.samplingrate * self.windowwidth))  # x axis times
        fig.set_size_inches(18, 10)
        fig.suptitle(f'{self.windowwidth}s window at {round(timepoint, 2)} seconds (Index {idx}; Sample {section})',
                     fontsize=25)
        axarr.set_xlabel('Time (s)', fontsize=20)

        # Get and screen peaks
        peaks, properties = _get_peaks(inputdata=normed_data, distance=self.mindistance, prominence=self.minamplitude,
                                       k=self.k)
        outliers, prom_thresholds, height_thresholds, _, _, ibi_outliers, baseheight = \
            _outlier_detect(inputdata=normed_data, inputpeaks=peaks, properties=properties,
                            minpeaks=self.numpeaksneeded, minwindow=self.minwindow, samplingrate=self.samplingrate,
                            mad=self.mad, ibimad=self.ibimad, bpmrange=self.bpmrange, rmssdrange=self.rmssdrange)

        # Plot Rapid HRV data
        axarr.set_ylabel('Processed data', fontsize=20)
        axarr.set_ylim(-5, 105)
        # axarr.set_xlim(self.exptime[0]-10, self.exptime[-1]+10)
        axarr.plot(segtime, normed_data, linewidth=5)  # Plot processed data
        axarr.plot(peaks / self.samplingrate, normed_data[peaks], marker="x", color="red", markersize=20,
                   linewidth=0)  # plot peaks

        if peaks.size > 1:  # Plot peak-relevant data if more than one peak
            axarr.plot(np.repeat(height_thresholds[0], self.windowwidth + 1), "--",
                       color="gray", linewidth=3)  # plot height thresholds
            axarr.plot(np.repeat(height_thresholds[1], self.windowwidth + 1), "--", color="gray",
                       linewidth=3)  # +1 to extend to end
            axarr.plot(peaks / self.samplingrate, prom_thresholds[0], "--", color="purple",
                       linewidth=3)  # plot height thresholds
            axarr.plot(peaks / self.samplingrate, prom_thresholds[1], "--", color="purple", linewidth=3)
            axarr.plot(peaks / self.samplingrate, baseheight, "--", color="black", linewidth=3)

        # Plot Labchart data
        text_box = f'HR = {np.round(self.bpm.iloc[idx], 2)}, RMSSD = {np.round(self.rmssd.iloc[idx], 2)}'
        if outliers:
            axarr.text(0, 0, text_box, fontsize=18, bbox=dict(boxstyle="round", facecolor="pink"))
        else:
            axarr.text(0, 0, text_box, fontsize=18, bbox=dict(boxstyle="round", facecolor="white"))

        fig.show()
